﻿:Class Parser
⍝ Parse a string and set switches values accordingly. DanB08 V1.19.
⍝ See Vector article in vol 19.4 for rules and logic behind.

⍝ This class has a few more features:
⍝ - you can specify the exact number of arguments or
⍝ - that the last argument includes all remaining text (LONG) and/or
⍝ - the maximum number (SHORT)
⍝ - the lowest generated error can be specified
⍝ - no space may be allowed before each switch (/a/la/DOS)
⍝ - switch names can be UPPER or mixcased. UPPER means they can be entered in any case but used UPPER case.
⍝ - use a prefix for variable names holding the switches value (e.g. /0 can go into variable X0)
⍝ - use " instead of ' (the other quote can then be used in between as in "I'm")
⍝ - there is a function to propagate the switches

⍝ To use these features call the class with a second string containing
⍝ 0 or more of 'nargs=n error=nnn allownospace upper prefix='

⍝ Example: create a parser accepting 3 switches in UPPERCASE: DATE which takes a value, SW2 which does
⍝ not take a value and SW3 which MAY take one of 'a', 'bc' or 'def'
⍝   ps←⎕NEW Parser ('/date= /sw2 /sw3[=]a bc def'  'upper')    ⍝ switches' variables are uppercase
⍝   data←ps.Parse 'dsa x /sw3=bc /dat="07/08/28"'   ⍝ will return a namespace containing
⍝   data.SW2≡0                                      ⍝ 2 strings and set 3 switches
⍝   data.SW3≡'bc'
⍝   '07/08/28'≡data.Switch 'DATE'                   ⍝ get the switch value from the table
⍝   '07/08/28'≡data.DATE                            ⍝ or directly from the NAME

⍝ If a one time parsing is needed the following will do:
⍝ data←(⎕new Parser '/sw1 /sw2...').Parse string

⍝ Example: create a parser accepting up to 4 arguments and one switch accepting vowels only:
⍝   pv←⎕new Parser ('-letter∊aeiou' 'nargs=4S')
⍝   words←pv.Parse '"I can''t"  ''word 2''  -let=aaaooo'  ⍝ only 2 args supplied
⍝   7 6≡⊃,/⍴¨words.Arguments

    ⎕io←0 ⋄ ⎕ml←1
    (LOWER UPPER)←'abcdefghijklmnopqrstuvwxyzàáâãåèéêëòóôõöøùúûäæü' 'ABCDEFGHIJKLMNOPQRSTUVWXYZÀÁÂÃÅÈÉÊËÒÓÔÕÖØÙÚÛÄÆÜ'
    ⎕fx's←upperCase s;b;i' '→0↓⍨1∊b←(⍴LOWER)>i←LOWER⍳s' '(b/s)←UPPER[b/i]'
    xCut←{⍺←⍵=1↑⍵ ⋄ 1↓¨⍺⊂⍵}     ⍝ exclusive cut
    fixCase←{⍵}                 ⍝ assume no fix
    sqz←{(b⍲1⌽b←' '=s)/s←' ',⍵} ⍝ remove double spaces, add leading ' '
    if←/⍨

    :field ERROR0←700           ⍝ errors are raised starting at this number
    :field DELIMITER            ⍝ the character used to delimit switches
    :field FORCESPACE           ⍝ whether a space MUST precede a switch
    :field NARGS←⍬              ⍝ how many arguments must be entered (⍬=any number)
    :field LONG←0               ⍝ does syntax support long scope?
    :field PREFIX←''            ⍝ switches' prefix

    ∇ init arg;model;features;b;f;na;mem;cut
⍝ Initialize class
      :Access public
      :Implements constructor
      (model features)←2↑⊂⍣(1≡≡arg)+arg
      DELIMITER←⍬⍴model,⊃⎕AV   ⍝ 1st char of 'model' delimits switches.
      FORCESPACE←' '≠DELIMITER ⍝ force only if NOT already used as delimiter
      :If ' '∨.≠features
          Ps←⎕NEW(⊃⊃⎕CLASS ⎕THIS)' nargs= upper allownospace error= prefix=' ⍝ allowed switches for the parser itself
          Pset←Ps.Parse 1↓sqz features
⍝ Find how many arguments this parser accepts.
⍝ This should be a ≥0 number either preceeded or followed by the letter L or S (for 'Long/Short' scope)
          :If 0<⍴NARGS←na←{(0≢⍵)/⍵}Pset.nargs  ⍝ unspecified?
              f←0<-/LONG←∨/2 2⍴'lLsS'∊na ⋄ (b NARGS)←⎕VFI na~'lLsS[]'
              'Invalid number of arguments'⎕SIGNAL 11 if(1≠⍴NARGS)∨(0∊b)∨f∧0∊NARGS
          :EndIf
          FORCESPACE←~Pset.allownospace
          fixCase←upperCase⍣Pset.upper          ⍝ allow starting errors in range 100-990
          'Error must be from 100 to 990'⎕SIGNAL 11 if 100 991=.≤ERROR0←700 Pset.Switch'error'
          'Prefix must be a valid name'⎕SIGNAL 11 if 0>⎕NC'x',⍨PREFIX←(b≢0)/⍕b←Pset.prefix
          ⎕EX'Ps' ⍝ no need anymore
      :EndIf
     
⍝ Parse the model
      cut←(Quotes model)<model∊1↑model
      SwTable←↑{2↑(¯1⌽1↓1,⍨<\⍵∊'[=∊')⊂⍵}¨cut xCut model
      SwTable[;1]←{deQuote(-⊥⍨' '=⍵)↓⍵}¨SwTable[;1]
      f←SwTable[;0]←fixCase¨SwTable[;0]~¨' '
      'switches must be unique'⎕SIGNAL 11 if f≢∪f
      'switches must be valid identifiers'⎕SIGNAL 11 if ¯1∊∊⎕NC¨PREFIX∘,¨f
      :If 1∊b←∨⌿mem←''∘⍴¨'∊' '[∊]'∘.⍷f←SwTable[;1] ⋄ (mem f)←b∘/¨mem f
          'A set must be provided with ∊'⎕SIGNAL 11 if 1∊(1 3+.×mem)=∊⍴¨f
      :EndIf
     
⍝ Create a namespace based on that model
⍝ (we could have created a similar class but we would have had to use ⎕NEW instead of ⎕NS ⎕OR)
      pData←{'#'∊⍕⎕THIS:#.⎕NS ⍵ ⋄ ⎕SE.⎕NS ⍵}''
     
      f←SwTable[;0]~'Switch' 'Propagate' ⍝ ensure fn names not hidden by switches
      ⍎(0<⍴f)/⍕'pData.(',(PREFIX∘,¨f),')←0'
      pData.⎕FX ⎕NR'Switch'    ⍝ do not use a ref to avoid keeping a copy of this ns
      pData.⎕FX ⎕NR'Propagate' ⍝ Mixed letter names minimize clashing with SWITCHES' names
      pData.(Delim ⎕IO ⎕ML)←DELIMITER 0 1
    ∇

⍝ This function is used to return a switch' value, possibly defaulted, e.g.
⍝ sw←123 Switch 'abc' ⍝ if 'abc' has not been set 123 is returned.
⍝ If /abc=789 was specified in the parsed string the value returned will be ,789, not '789'

    ∇ r←{def}Switch s;⎕IO;v
⍝ Return switch' value
      ⎕IO←r←0 ⍝ invalid switches are considered not there
      :Trap 3
          r←1⊃SwD[SwD[;0]⍳⊂,s;]
      :EndTrap
      :If 0≠⎕NC'def' ⍝ even undefined switches can be defaulted
          :If 0≡r ⋄ r←def ⍝ use default if not set
          :ElseIf (1≢r)∧2|⎕DR,def ⍝ num is 11, x3, 645, 1287, 1289
              r←1⊃v←⎕VFI r ⍝ make numeric if default is also numeric
              ('value must be numeric for ',s)⎕SIGNAL 11↓⍨∧/0⊃v ⍝ <if> unavailable
          :EndIf
      :EndIf
    ∇

    ∇ mask←Quotes str;Q;qm;tq;n
⍝ Find where text delimited by ' or " starts/ends
      qm←str∊'''"' ⋄ n←⍴Q←qm/str   ⍝ work on quotes only
      tq←<\(∘.<⍨⍳n)∧∘.=⍨Q          ⍝ matching trailing markers
      mask←n⍴tq⌹(∘.=⍨⍳n)-0 ¯1↓0,tq ⍝ all trailing markers
      mask←qm\mask∨¯1↓1,mask       ⍝ lead & trail markers
    ∇

    ∇ str←deQuote str;m;lq
⍝ Remove quotes and double quotes
      →0↓⍨∨/m←str∊'''"'            ⍝ which quote to use
      lq←m<1↓1,⍨m←str≠str[m⍳1]     ⍝ last quote
      str←(lq<m∨=\m)/str           ⍝ those to keep
    ∇

⍝ Parsing function.
⍝ Spaces are used to delimit arguments.
⍝ Quotes must be used to include spaces or delimiter in arguments.
⍝ Switches can exist with or without [possibly defaulted] value
⍝ or be elided. Switches not mentioned are refused.
⍝ Shorter names are accepted but '=' MUST be used to supply values.

    ∇ data←Parse arg;Er;s;t;sw;table;parms;swit;swmat;m;val;vnc;set;req;pat;nov;q;∆;i;p;bad;twv;Q
⍝ PARSE switches and reset argument. Version 3.2
      :Access public
⍝ Account for "
      Er←ERROR0 ⋄ s←-FORCESPACE ⋄ arg←' ',arg
      t←≠\Q←Quotes arg ⋄ s←∨\sw←s⌽t<arg⍷⍨(s-1)↑DELIMITER ⍝ QUOTED ⋄ all SWITCHES
      'unbalanced quotes'⎕SIGNAL Er if ¯1↑t
      data←{'#'∊⍕⎕THIS:#.⎕NS ⍵ ⋄ ⎕SE.⎕NS ⍵}⎕OR'pData' ⋄ table←SwTable[;,0],0
⍝ Split the arguments from the switches
      parms←(~s)/arg
      swit←{s←{(-⊥⍨' '=⍵)↓⍵}⍵ ⋄ (⍴s)≠⍴n←fixCase s↑⍨s⍳'=':n((1+⍴n)↓s) ⋄ n 1}¨sw xCut arg
      swmat←' ',⍨↑SwTable[;0]
     
⍝ Process each switch separately according to their position
      :While 0<⍴swit
          (sw val)←⊃swit
          :If 1≠+/i←((sw,' ')⍷swmat)[;0] ⍝ exact matches
          :AndIf 1≠t←+/i←(sw⍷swmat)[;0]
              m←((×t)⊃'unknown' 'ambiguous'),' switch: <',sw,'>'
              m ⎕SIGNAL Er+1
          :EndIf
          vnc←'['=⊃1⊃sw pat←,i⌿SwTable ⍝ '[=]' means 'value not compulsory'
⍝ 'pat' is 1 switch group complete with assignment and allowed values if present
          set←'∊'=⊃vnc↓pat ⋄ pat[set/vnc]←'=' ⍝ is the right hand string a set?
⍝ We need to check if a value is needed or not allowed
          req←'='=⊃vnc↓pat ⋄ nov←(0∊⍴val)∨twv←1≡val ⍝ required ⋄ no value set ∨ there w/o value
          m←1⌽'>no value allowed for switch <',sw
          m ⎕SIGNAL(req∨twv)↓Er+2
          m←1⌽'>value required for switch <',sw
          m ⎕SIGNAL(nov∧req>vnc)⍴Er+3
          table[i⍳1;1]←⊂val←deQuote val ⍝ remove extra quotes
⍝ Valid strings are supplied with the switch: verify them if a value supplied
          :If twv<0<⍴p←(1+vnc+pat⍳'=')↓pat  ⍝ an arg?
⍝ Spaces are invalid in args
              bad←(set∧∧/val∊p)⍱set<(' '∊val)<∨/(' ',val,' ')⍷' ',p,' '
              m←'invalid value for switch <',sw,'> (must be ',(set⊃'ONE of' 'ALL in'),' "',p,'")'
              m ⎕SIGNAL bad⍴Er+4
          :EndIf
⍝ Assign in instance
          :Trap 2 ⋄ ⍎'data.',PREFIX,sw,'←val' ⋄ :EndTrap
          swit←1↓swit
      :EndWhile
     
      data.SwD←table
     
⍝ Find each token in the argument and use the delimiter to separate them.
⍝ The following allows us to deal with '' properly:
      s←parms∊' ' ⋄ ∆←≠\q←(⍴parms)↑Q ⋄ i←q∧1⌽q ⋄ t←q>¯1⌽q ⍝ i=double quote, t=1st quote
      p←t∨s>∆ ⍝ DELIMITERS: spaces NOT in quotes OR 1st quote
⍝ COMPRESSOR: any double quote OR last quote NOT in ≠\ or double delim.
      t←(∆<i∨q>1⌽q)⍱p∧1⌽p
      s←+/p←t/p ⋄ q←~p ⍝∇⋄
      parms←q\q/t/parms ⍝ turn quote delimiters into spaces for Long below
     
⍝ Make sure the number of arguments matches the required number
      :If 0<⍴NARGS ⋄ :AndIf 0≠m←×s-NARGS ⋄ t←'too ',((m=1)⊃'few' 'many'),' arguments'
          t ⎕SIGNAL Er+5 if(m≠-/LONG)∧⍲/LONG ⍝ accept both Long and Short
          p←p∧NARGS≥+\p
      :EndIf
      data.Arguments←p xCut parms
    ∇

    ∇ str←{Del}Propagate switches;sw;b;v;si
⍝ Recreate a string of the switches in order to be passed to another command
⍝ e.g. myArgs.Propagate 'VERSION'
⍝ Invalid switches are ignored.
      str←'' ⋄ si←SwD[;0]∊sw←({⍺←⍵=1↑⍵ ⋄ 1↓¨⍺⊂⍵}' ',switches)~⊂''
      :If 1∊b←0≢¨v←si/SwD[;1]
          :If 0=⎕NC'Del' ⋄ Del←Delim ⋄ :EndIf
          str←∊(b/v){' ',Del,⍵,(1≢⍺)/'=',{∨/⍵∊Del,' ''"':Q,((1+⍵=Q)/⍵),Q←'"' ⋄ ⍵}⍕⍺}¨b/si/SwD[;0]
      :EndIf
    ∇

:EndClass